/***********      .---.         .-"-.      *******************\
* -------- *     /   ._.       /  ` \     * ---------------- *
* Author's *     \_  (__\      \_v_/     * humus@rogers.com *
*   note   *     //   \\       //   \\     * ICQ #47010716    *
* -------- *    ((     ))     ((     ))    * ---------------- *
*          ****--""---""-------""---""--****                  ********\
* This file is a part of the work done by Humus. You are free to use  *
* the code in any way you like, modified, unmodified or copy'n'pasted *
* into your own work. However, I expect you to respect these points:  *
*  @ If you use this file and its contents unmodified, or use a major *
*    part of this file, please credit the author and leave this note. *
*  @ For use in anything commercial, please request my approval.      *
*  @ Share your work and ideas too as much as you can.                *
\*********************************************************************/

#include "App.h"

BaseApp *app = new App();

void App::moveCamera(const vec3 &dir){
	vec3 newPos = camPos + dir * (speed * frameTime);

	// Collision detection
	vec3 point;
	const BTri *tri;
	if (bsp.intersects(camPos, newPos, &point, &tri)){
		newPos = point + tri->plane.xyz();
	}
	bsp.pushSphere(newPos, 20);

	camPos = newPos;
}

void App::resetCamera(){
	camPos = vec3(-384, 812, 0);
	wx = 0.715f;
	wy = PI * 1.5f;
}

bool App::init(){
	lightPos = vec3(0, 0, 0);

	map = new Model();
	if (!map->loadT3d("../Models/Corridor/Map.t3d")) return false;
	map->scale(0, vec3(-1, 1, 1));
	map->removeStream(4);
	map->removeStream(3);
	map->removeStream(2);
	map->computeTangentSpace(true);

	uint nIndices = map->getIndexCount();
	vec3 *src = (vec3 *) map->getStream(0).vertices;
	uint *inds = map->getStream(0).indices;

	for (uint i = 0; i < nIndices; i += 6){
		vec3 v0 = src[inds[i]];
		vec3 v1 = src[inds[i + 1]];
		vec3 v2 = src[inds[i + 2]];
		vec3 v3 = src[inds[i + 5]];

		bsp.addTriangle(v0, v1, v2);
		bsp.addTriangle(v0, v2, v3);
	}
	bsp.build();

	map->cleanUp();
	map->changeAllGeneric(true);


	// Create volume lightmap if missing
#define SIZE 128
	
	vec3 min, max;
	map->getBoundingBox(0, min, max);

	vec3 diff = max - min;

	scale = 1.0f / diff;
	bias = -min * scale;

	if (!lmImage.loadDDS("LightMap.dds")){
		ubyte *dst = lmImage.create(FORMAT_I8, SIZE, SIZE, SIZE, 1);

		vec3 samples[255];
		samples[0] = vec3(0, 0, 0);
		for (int i = 1; i < 255; i++){
			vec3 d;
			do {
				d.x = float(rand()) / RAND_MAX * 2.0f - 1.0f;
				d.y = float(rand()) / RAND_MAX * 2.0f - 1.0f;
				d.z = float(rand()) / RAND_MAX * 2.0f - 1.0f;
			} while (dot(d, d) > 1.0f);
			samples[i] = d * 50;
		}

		for (int z = 0; z < SIZE; z++){
			for (int y = 0; y < SIZE; y++){
				for (int x = 0; x < SIZE; x++){
					vec3 pos = min + vec3(float(x), float(y), float(z)) / float(SIZE - 1) * diff;

					bsp.pushSphere(pos, 9.9f);
					bsp.pushSphere(pos, 9.8f);
					bsp.pushSphere(pos, 9.7f);

					ubyte sum = 0;
					if (bsp.isInOpenSpace(pos)){
						for (int i = 0; i < 255; i++){
							if (!bsp.intersectsCached(pos, lightPos + samples[i])) sum++;
						}
					}

					*dst++ = sum;
				}
			}
		}
		lmImage.saveDDS("LightMap.dds");
	}

	return true;
}

void App::exit(){
	delete map;
}

bool App::load(){
	if (!GLSL_supported){
		ErrorMsg("No GLSL support");
		return false;
	}

	multiPass = false;
	const char *vendor = (const char *) glGetString(GL_VENDOR);
	if (strstr(vendor, "ATI Technologies")){
		int indirections = 0;
		glGetProgramivARB(GL_FRAGMENT_PROGRAM_ARB, GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB, &indirections);

		// Make R3xx and R4xx cards use multipass
		multiPass = (indirections <= 4);
	}

	// Shaders
	const char *attribs[] = { NULL, "textureCoord", "tangent", "binormal", "normal" };
	if (multiPass){
		if ((lighting[0] = renderer->addShader("lighting.shd", attribs, elementsOf(attribs), "#define COUNT 10\n#define FIRST_PASS\n")) == SHADER_NONE) return false;
		if ((lighting[1] = renderer->addShader("lighting.shd", attribs, elementsOf(attribs), "#define COUNT 15\n")) == SHADER_NONE) return false;
	} else {
		if ((lighting[0] = renderer->addShader("lighting.shd", attribs, elementsOf(attribs), "#define COUNT 40\n#define FIRST_PASS\n")) == SHADER_NONE) return false;
	}
	if ((preZ = renderer->addShader("preZ.shd")) == SHADER_NONE) return false;


	// Samplerstates
	if ((linear = renderer->addSamplerState(LINEAR, WRAP, WRAP, WRAP)) == SS_NONE) return false;
	if ((trilinearAniso = renderer->addSamplerState(TRILINEAR_ANISO, WRAP, WRAP, WRAP)) == SS_NONE) return false;
	
	// Textures
	if ((base[0] = renderer->addTexture("../Textures/floor_wood_3.dds", true, trilinearAniso)) == TEXTURE_NONE) return false;
	if ((base[1] = renderer->addTexture("../Textures/stone34.dds",      true, trilinearAniso)) == TEXTURE_NONE) return false;
	if ((base[2] = renderer->addTexture("../Textures/floor_wood_4.dds", true, trilinearAniso)) == TEXTURE_NONE) return false;

	if ((bump[0] = renderer->addNormalMap("../Textures/floor_wood_3Bump.dds", FORMAT_RGBA8, true, trilinearAniso)) == TEXTURE_NONE) return false;
	if ((bump[1] = renderer->addNormalMap("../Textures/stone34Bump.dds",      FORMAT_RGBA8, true, trilinearAniso)) == TEXTURE_NONE) return false;
	if ((bump[2] = renderer->addNormalMap("../Textures/floor_wood_4Bump.dds", FORMAT_RGBA8, true, trilinearAniso)) == TEXTURE_NONE) return false;

	if ((noColorWrite = renderer->addBlendState(ONE, ZERO, BM_ADD, NONE)) == BS_NONE) return false;
	if ((oneSrcAlpha = renderer->addBlendState(ONE, SRC_ALPHA, BM_ADD, ALL)) == BS_NONE) return false;
	

	// Generate a noise texture fog the fog
#define SCALE 0.18293f

	Image img;
	if (!img.loadDDS("Noise.dds")){
		initNoise();

		ubyte *dest = img.create(FORMAT_I8, 64, 64, 64, 1);

		for (int z = 0; z < 64; z++){
			for (int y = 0; y < 64; y++){
				for (int x = 0; x < 64; x++){
					float n = tileableTurbulence3(x * SCALE, y * SCALE, z * SCALE, 64 * SCALE, 64 * SCALE, 64 * SCALE, 8.0f);
					n = saturate(n * 1.0f + 0.5f);

					*dest++ = (ubyte) (n * 255.0f + 0.5f);
				}
			}
		}
		img.saveDDS("Noise.dds");
	}
	if ((noise = renderer->addTexture(img, linear)) == TEXTURE_NONE) return false;

	// A volume lightmap
	if ((lightMap = renderer->addTexture(lmImage, linearClamp)) == TEXTURE_NONE) return false;

	// Upload map to vertex/index buffer
	if (!map->makeDrawable(renderer, true)) return false;

	return true;
}

void App::unload(){

}

void App::drawFrame(){
	mat4 projection = perspectiveMatrixX(1.5f, width, height, 10, 2500);
	mat4 modelview = rotateXY(-wx, -wy) * translate(-camPos);
	mat4 mvp = projection * modelview;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	// Pre-Z pass to speed things up
	renderer->reset();
	renderer->setShader(preZ);
	renderer->setShaderConstant4x4f("mvp", mvp);
	renderer->setRasterizerState(cullBack);
	renderer->setBlendState(noColorWrite);
	renderer->apply();


	map->draw(renderer);


	int passCount = multiPass? 3 : 1;

	// Main lighting and fogging pass
	for (int p = 0; p < passCount; p++){
		for (uint i = 0; i < map->getBatchCount(); i++){
			renderer->reset();
			renderer->setRasterizerState(cullBack);
			if (p == 0){
				renderer->setShader(lighting[0]);
			} else {
				renderer->setShader(lighting[1]);
				renderer->setBlendState(oneSrcAlpha);
			}
			renderer->setTexture("Base", base[i], trilinearAniso);
			renderer->setTexture("Bump", bump[i], trilinearAniso);
			renderer->setTexture("Fog", noise, linear);
			renderer->setTexture("LightMap", lightMap, linearClamp);

			renderer->setShaderConstant4x4f("mvp", mvp);
			renderer->setShaderConstant3f("camPos", camPos);

			renderer->setShaderConstant3f("shadowScale", scale);
			renderer->setShaderConstant3f("shadowBias", bias);

			vec4 fogScaleBias(0, -0.06f * time, 0, 0.0008f);
			renderer->setShaderConstant4f("fogScaleBias", fogScaleBias);

			if (p > 0){
				float nStep = float(15 * p - 5);
				renderer->setShaderConstant1f("fogStart", nStep * fogScaleBias.w);
				renderer->setShaderConstant3f("shadowStart", nStep * scale);
			} else {
				renderer->setShaderConstant3f("lightPos", lightPos);
			}

			renderer->apply();

			map->drawBatch(renderer, i);
		}
	}
}
